package com.maenle.bump.util

import java.util.Base64
import javax.crypto.spec.PBEKeySpec
import javax.crypto.SecretKeyFactory
import com.macasaet.fernet.Key
import com.macasaet.fernet.Token
import com.macasaet.fernet.StringValidator
import com.macasaet.fernet.Validator
import com.maenle.bump.ui.MainActivity
import java.math.BigInteger
import java.security.SecureRandom
import java.time.Duration
import java.time.temporal.TemporalAmount

class MessageProcessor(code: String, private val salt: ByteArray? = null) {
    private var pSender: String
    private var pPassword: String

    init {
        codeValid(code).let {
            val codeSplit: List<String> = code.split("-")
            pSender = codeSplit.subList(0, SENDER_LENGTH).joinToString("-")
            pPassword = codeSplit.subList(SENDER_LENGTH, codeSplit.size).joinToString("-")
        }
    }

    val password: String
    get() {
        return pPassword
    }

    val sender: String
        get() {
            return pSender
        }

    val code: String
        get() {
            return "$pSender-$pPassword"
        }

    private fun codeValid(code: String): Boolean {
        return code.split("-").size >= KEY_LENGTH + SENDER_LENGTH
    }

    fun encrypt(decrypted: String): String {
        val message = MessageEncrypt(decrypted)
        salt?.let { message.updateSalt(salt) }
        return message.encryptWith(pPassword)
    }

    fun decrypt(messageRaw: String): String {
        val message = MessageDecrypt(messageRaw)
        return message.validateAndDecryptWith(pPassword)
    }


    private class MessageEncrypt(val messageDecrypted: String) : Message() {
        fun encryptWith(password: String): String {
            val fernetKey = deriveMessageKeyFromPassword(password)
            val token = Token.generate(fernetKey, messageDecrypted)
            val tokenString = Base64.getUrlDecoder().decode(token.serialise())

            val iterationBuffer = ByteArray(4)
            for (i in 0..3) iterationBuffer[i] = (iterations shr (i*8)).toByte()
            iterationBuffer.reverse()

            val data = salt + iterationBuffer + tokenString
            return String(Base64.getUrlEncoder().encode(data))
        }
    }


    private class MessageDecrypt(messageRaw: String) : Message() {
        init {
            super.splitMessage(messageRaw)
        }
        fun validateAndDecryptWith(password: String, timeToLive: Long = TIME_TO_LIVE): String {
            val fernetKey = deriveMessageKeyFromPassword(password)
            val validator: Validator<String> = object : StringValidator {
                override fun getTimeToLive(): TemporalAmount {
                    return Duration.ofHours(timeToLive)
                }
            }
            return token.validateAndDecrypt(fernetKey, validator)
        }
    }


    abstract class Message(var salt: ByteArray = freshSalt, var iterations: Int = ITERATIONS) {
        lateinit var token: Token

        protected fun splitMessage(message : String) {
            val decodedMessage : ByteArray = Base64.getUrlDecoder().decode(message)

            salt = decodedMessage.copyOfRange(0, SALT_LENGTH)
            val iterationsDecoded = decodedMessage.copyOfRange(SALT_LENGTH, SALT_LENGTH + ITERATIONS_LENGTH)
            val tokenString = String(Base64.getUrlEncoder().encode(decodedMessage.copyOfRange(20, decodedMessage.size)))

            iterations = BigInteger(iterationsDecoded).toInt()
            token = Token.fromString(tokenString)
        }

        override fun equals(other: Any?): Boolean {
            if (this === other) return true
            if (javaClass != other?.javaClass) return false

            other as Message

            if (!salt.contentEquals(other.salt)) return false
            if (iterations != other.iterations) return false
            if (token != other.token) return false

            return true
        }

        override fun hashCode(): Int {
            var result = salt.contentHashCode()
            result = 31 * result + iterations
            result = 31 * result + token.hashCode()
            return result
        }

        fun updateSalt(newSalt: ByteArray) {

            if(newSalt.size == SALT_LENGTH) {
                salt = newSalt
            }
        }

        companion object {
            val freshSalt: ByteArray
                get()
                {
                    val salt = ByteArray(16)
                    SecureRandom().nextBytes(salt)
                    return salt
                }
        }

        fun deriveMessageKeyFromPassword(password: String): Key {
            val derivedKeyLength = 256
            val spec = PBEKeySpec(password.toCharArray(), salt, iterations, derivedKeyLength)
            val secretKeyFactory = SecretKeyFactory.getInstance("PBKDF2WithHmacSHA256")
            val key = secretKeyFactory.generateSecret(spec).encoded
            val saltedKey = Base64.getUrlEncoder().encodeToString(key)

            return Key(saltedKey)
        }
    }

    @Suppress("unused")
    companion object {
        private val TAG = MainActivity::class.java.simpleName
        const val KEY_LENGTH: Int = 8
        const val SENDER_LENGTH: Int = 4
        const val SALT_LENGTH: Int = 16
        const val ITERATIONS_LENGTH: Int = 4

        const val ITERATIONS: Int = 100_000
        const val TIME_TO_LIVE: Long = 100000 * 365 * 24

    }


}